/**
  Funky CPU emulator.
  Orion Sky Lawlor, olawlor@acm.org, 2005/09/12 (Public Domain)
*/
#include <stdio.h>
#include <stdlib.h>

FILE *f; /* Input file we'll read from */

/**
Quit the emulator.
*/
void quit(const char *why,int value1=0,int value2=0) {
	fprintf(stderr,why,value1,value2);
#ifdef WIN32
	printf("\nPress return to exit...\n");
	getc(stdin);
#endif
	printf("\n");
	exit(0);
}

/** A tiny CPU Emulator */
class CPU {
public:
	enum {n_regs=8};
	signed int regs[n_regs]; ///< "registers"-- temporary working space of machine
	
	enum {n_mem=1024};
	unsigned char mem[n_mem]; ///< "memory"-- just program memory for now
	
	int pc; ///< "Program counter": location in memory we're running now.
	int lessflag; ///< Less-than flag (set by compare instruction)
	
	int run(int verbose);
	void dump(int id); /* Print out registers */
};

/*************** Main Emulator Run Routine *************/
int CPU::run(int verbose) {
	//return ((int(*)())&mem)();
	while (1) {
		/* fetch instruction */
		int opcode=mem[pc++]; /* First byte tells you what's happening */
		int next=mem[pc++]; /* next byte tells you who it's happening to */
		int dest=(next&070)>>3, src=next&007; /* destination and source registers */
		if (verbose) printf("   pc=%d: opcode=0x%02x next=%d, 0%03o (dest==%d, src=%d)\n",
		                         pc-2,opcode,next,next,dest,src);
		/* decode instruction */
		switch (opcode) { 
		case 0xC3: return regs[0]; break; /* RET--return a value to outside program */
		/* Data moves */
		case 0x8B: regs[dest] = regs[src]; break; /* MOV */
		case 0xB0: case 0xB1: case 0xB2: case 0xB3: /* MOV immediate (0xB0+dest) */
			regs[opcode&007] = next; break;
		/* Arithmetic */
		case 0x03: regs[dest] += regs[src]; break; /* ADD */
		case 0x2B: regs[dest] -= regs[src]; break; /* SUB */
		case 0x23: regs[dest] &= regs[src]; break; /* AND */
		case 0x09: regs[dest] |= regs[src]; break; /* OR */
		case 0x33: regs[dest] ^= regs[src]; break; /* XOR */
		/* Control */
		case 0x3B: lessflag = regs[dest]<regs[src]; break; /* CMP: compare */
		case 0x7C: if (lessflag) pc+=(signed char)next; break; /* JL: jump-if-less */
		case 0xEB: pc+=(signed char)next; break; /* JMP */
		/* Special instructions */
		case 0x90: break; /* NOP-- do nothing! */
		case 0xFF:
			if (next!=0x57) quit("Bad 0xFF instruction %02x at pc %d\n",pc-3);
			else switch (mem[pc++]) {
				case 0x00: dump(pc-3); break;
				case 0x10: printf("Register 0 is %d (0x%08X)\n",regs[0],regs[0]); break;
				case 0x20: fscanf(f,"%i",&regs[0]); break;
				default: quit("Unrecognized special 0xff function at pc %d\n",pc-3); break;
			}; 
			break;
		default: quit("Unrecognized opcode 0x%x at pc %d\n",opcode,pc-2); break;
		}
	}
}

void CPU::dump(int id) {
	printf("Dump at pc %d\n",id);
	for (int i=0;i<8;i++)
		if (regs[i]!=0)
			printf("Register[%X]=0x%X  (or %d)\n",i,regs[i],regs[i]);
	printf("\n");
}

int main(int argc,char *argv[]) {
	CPU c={0};
	c.pc=0; /* Start at first instruction of program */
	int verbose=0;
	
/* Read program to run from the input file */
	const char *inFile="input.txt"; if (argc>1) inFile=argv[1];
	f=fopen(inFile,"r"); if (stdin==NULL) quit("Can't open input file");
	char inLine[1024];
	int loadCount=0;
	while (1) { /* Keep reading until something goes wrong */
		int val; 
		/* Try to read numbers */
		while (1==fscanf(f,"%i",&val)) {
			if (val<0) {/* negative number marks end of input */
				if (val==-2) verbose=1; /* -2 marks end of input, and turns on verbose mode */
				goto done_reading; 
			}
			else { /* Just a normal program instruction */
				if (val>255) {
					printf("Loaded value %d from input file at pc=%d--can't fit into a byte!  Check your input file.\n",
						val,loadCount);
					exit(1);
				}
				c.mem[loadCount++]=val;
			}
		}
		if (feof(stdin)) goto done_reading;
		/* Can't parse that--probably a comment.  Skip the whole line. */
		fgets(inLine,sizeof(inLine),f);
	}
done_reading:
	
	/* Run the program */
	int v=c.run(verbose);
	printf("Program complete.  Return code: %d  (0x%X)\n",v,v);
	
	if (verbose) {
	/* Print out the loaded program (to make sure it loaded right!) */
		printf("Loaded program contents (%d bytes total)\n",loadCount);
		for (int pc=0;pc<loadCount;pc++) {
			printf(" pc = %3d: 0x%02x = 0%03o = %3d\n",pc,c.mem[pc],c.mem[pc],c.mem[pc]);
		}
	}
	return 0;
}
